@@ -3,7 +3,7 @@ Record generation and templating
 """
 import logging
 from collections import ChainMap, OrderedDict
-from typing import Optional
+from typing import Dict, List, Optional, Set
 
 import jinja2
 
@@ -12,7 +12,13 @@ import pytmc
 from .default_settings.unified_ordered_field_list import unified_lookup_list
 
 logger = logging.getLogger(__name__)
-MAX_ARCHIVE_ELEMENTS = 1024
+
+# NOTE: Some PLCs may fill a buffer on a per-cycle basis. At 1KHz, 1000 is the
+# magic number that would be used.  This threshold reflects the fact that we do
+# not want to include those PVs in the archiver:
+MAX_ARCHIVE_ELEMENTS = 999
+
+# This field length maximum is based on the .DESC field length in epics-base:
 MAX_DESC_FIELD_LENGTH = 40
 
 _default_jinja_env = jinja2.Environment(
@@ -119,6 +125,19 @@ class EPICSRecord:
                     fields = set(config[config_key].split(' '))
                     self.autosave[record_key] = fields
 
+    def update_pini_for_autosave(self):
+        """
+        Set PINI=1 if VAL is autosaved at pass1.
+
+        This is a bandaid fix for unexplained behavior where the
+        autosave pass1 does not send values to the PLC unless
+        PINI=1.
+
+        This is intended to be called for output records.
+        """
+        if 'VAL' in self.autosave['pass1']:
+            self.fields['PINI'] = 1
+
     def render(self, sort: bool = True):
         """Render the provided template"""
         for field, value in list(self.fields.items()):
@@ -165,6 +184,16 @@ class RecordPackage:
 
         self._parse_config(self.config)
 
+    def __repr__(self):
+        items = {
+            attr: getattr(self, attr, None)
+            for attr in ('tcname', 'pvname', 'long_description',
+                         'linked_to_pv')
+        }
+        info = ', '.join(f'{k}={v!r}' for k, v in items.items()
+                         if v)
+        return f"{self.__class__.__name__}({info})"
+
     def _parse_config(self, config):
         'Parse the chain configuration'
         if config is None:
@@ -185,7 +214,7 @@ class RecordPackage:
         return config
 
     def _configure_link(self):
-        'Link this record to a pre-existing EPICS record via a CA (CPP) link'
+        'Link this record to a pre-existing EPICS record via a CA (CP) link'
         self.linked_to_pv = self.config.get('link') or None
 
     def _configure_pvname(self):
@@ -277,9 +306,76 @@ class RecordPackage:
                     f'Unsupported data type {data_type.name} in chain: '
                     f'{chain.tcname} record: {chain.pvname}'
                 ) from None
+            if spec is IntegerRecordPackage:
+                if "scale" in chain.config or "offset" in chain.config:
+                    # longin/longout do not support scaling. Special-case and
+                    # cast to a FloatRecordPackage.
+                    spec = FloatRecordPackage
+
         return spec(*args, chain=chain, **kwargs)
 
 
+AutosaveDefaults = Dict[str, Dict[str, Set[str]]]
+
+
+def make_autosave_defaults(
+    input_pass0: Optional[List[str]] = None,
+    input_pass1: Optional[List[str]] = None,
+    output_pass0: Optional[List[str]] = None,
+    output_pass1: Optional[List[str]] = None,
+    exclude_defaults: bool = False,
+) -> AutosaveDefaults:
+    """
+    Create autosave defaults for a given record type.
+
+    Parameters
+    ----------
+    input_pass0 : List[str]
+        Fields to save on the input record for pass 0 (pre-iocInit)
+    input_pass1 : List[str]
+        Fields to save on the input record for pass 1 (post-iocInit)
+    output_pass0 : List[str]
+        Fields to save on the output record for pass 0 (pre-iocInit)
+    output_pass1 : List[str]
+        Fields to save on the output record for pass 1 (post-iocInit)
+    exclude_defaults : bool, optional
+        Exclude the defaults normally used for all records.
+
+    Returns
+    -------
+    dict
+        Dictionary of defaults with keys "input" and "output", under which
+        are keys "pass0" and "pass1".
+    """
+
+    if not exclude_defaults:
+        default_input_pass0 = {
+            "DISS",  # Disable Alarm Sevrty
+            "UDFS",  # Undefined Alarm Sevrty
+            "DESC",  # Description
+        }
+        default_output_pass0 = {
+            "VAL",  # The value (pass 0, does not cause a write)
+            "DISS",  # Disable Alarm Sevrty
+            "UDFS",  # Undefined Alarm Sevrty
+            "DESC",  # Description
+        }
+    else:
+        default_input_pass0 = set()
+        default_output_pass0 = set()
+
+    return dict(
+        input=dict(
+            pass0=set(input_pass0 or []) | default_input_pass0,
+            pass1=set(input_pass1 or []),
+        ),
+        output=dict(
+            pass0=set(output_pass0 or []) | default_output_pass0,
+            pass1=set(output_pass1 or []),
+        ),
+    )
+
+
 class TwincatTypeRecordPackage(RecordPackage):
     """
     The common parent for all RecordPackages for basic Twincat types
@@ -308,12 +404,7 @@ class TwincatTypeRecordPackage(RecordPackage):
        not required.
     """
     field_defaults = {'PINI': 1, 'TSE': -2}
-    autosave_defaults = {
-        'input': dict(pass0={},
-                      pass1={}),
-        'output': dict(pass0={'VAL'},
-                       pass1={}),
-    }
+    autosave_defaults = make_autosave_defaults()
 
     dtyp = NotImplemented
     input_rtyp = NotImplemented
@@ -322,6 +413,9 @@ class TwincatTypeRecordPackage(RecordPackage):
     output_only_fields = {'DOL', 'IVOA', 'IVOV', 'OMSL'}
     archive_fields = ['VAL']
 
+    # Is an auxiliary record required to support existing record linking?
+    link_requires_record = False
+
     def __init_subclass__(cls, **kwargs):
         """Magic to have field_defaults be the combination of hierarchy"""
         super().__init_subclass__(**kwargs)
@@ -408,9 +502,6 @@ class TwincatTypeRecordPackage(RecordPackage):
                              package=self,
                              )
 
-        # Set a default description to the tcname
-        _update_description(record, self.chain.tcname)
-
         # Add our port
         record.fields['INP'] = self.asyn_input_port_spec
         record.fields['DTYP'] = self.dtyp
@@ -423,6 +514,9 @@ class TwincatTypeRecordPackage(RecordPackage):
              }
         )
 
+        # Set a default description to the tcname
+        _update_description(record, self.chain.tcname)
+
         # Records must always be I/O Intr, regardless of the pragma:
         record.fields['SCAN'] = 'I/O Intr'
 
@@ -430,6 +524,35 @@ class TwincatTypeRecordPackage(RecordPackage):
 
         return record
 
+    def _get_omsl_fields(self):
+        """Get output mode select fields for the output record."""
+        if not self.linked_to_pv or self.linked_to_pv[-1] is None:
+            return {}
+
+        last_link = self.linked_to_pv[-1]
+        if last_link.startswith('*'):
+            # NOTE: A special, undocumented syntax for a lack of a better
+            # idea/more time:  need to allow pytmc to get access to a PV name
+            # it generates
+            # Consider this temporary API, only to be used in
+            # lcls-twincat-general for now.
+            pv_parts = list(self.config['pv'])
+            linked_to_pv = self.delimiter.join(
+                pv_parts[:-1] + [last_link.lstrip('*')]
+            )
+        else:
+            linked_to_pv = ''.join(
+                part for part in self.linked_to_pv
+                if part is not None
+            )
+
+        linked_to_pv = linked_to_pv.replace(self.macro_character, '$')
+        return {
+            'OMSL': 'closed_loop',
+            'DOL': linked_to_pv + ' CP MS',
+            'SCAN': self.config.get('link_scan', '.5 second'),
+        }
+
     def generate_output_record(self):
         """
         Generate the record to write values back to the PLC
@@ -452,37 +575,22 @@ class TwincatTypeRecordPackage(RecordPackage):
                              package=self,
                              )
 
-        # Set a default description to the tcname
-        _update_description(record, self.chain.tcname)
-
         # Add our port
         record.fields['DTYP'] = self.dtyp
         record.fields['OUT'] = self.asyn_output_port_spec
 
+        # By default, NO_ALARM if the output record has not processed yet:
+        record.fields.setdefault("UDFS", "0")
+
         # Remove timestamp source and process-on-init for output records:
         record.fields.pop('TSE', None)
         record.fields.pop('PINI', None)
 
-        if self.linked_to_pv and self.linked_to_pv[-1] is not None:
-
-            record.fields['OMSL'] = 'closed_loop'
-
-            last_link = self.linked_to_pv[-1]
-            if last_link.startswith('*'):
-                # NOTE: A special, undocumented syntax for a lack of a better
-                # idea/more time:  need to allow pytmc to get access to a PV
-                # name it generates
-                # Consider this temporary API, only to be used in
-                # lcls-twincat-general for now.
-                pv_parts = list(self.config['pv'])
-                linked_to_pv = ':'.join(pv_parts[:-1] +
-                                        [last_link.lstrip('*')])
-            else:
-                linked_to_pv = ''.join([part for part in self.linked_to_pv
-                                        if part is not None])
-
-            record.fields['DOL'] = linked_to_pv + ' CPP MS'
-            record.fields['SCAN'] = self.config.get('link_scan', '.5 second')
+        # Add on OMSL fields, if this is linked to an existing record.
+        # Some links (such as strings) may require auxiliary records, so
+        # don't repurpose the output record in that case.
+        if not self.link_requires_record:
+            record.fields.update(self._get_omsl_fields())
 
         # Update with given pragma fields - ignoring input-only fields:
         user_fields = self.config.get('field', {})
@@ -492,7 +600,11 @@ class TwincatTypeRecordPackage(RecordPackage):
              }
         )
 
+        # Set a default description to the tcname
+        _update_description(record, self.chain.tcname)
+
         record.update_autosave_from_pragma(self.config)
+        record.update_pini_for_autosave()
         return record
 
     @property
@@ -514,6 +626,20 @@ class BinaryRecordPackage(TwincatTypeRecordPackage):
                           'OUT', 'RBV', 'RPVT', 'WDPT'}
     input_only_fields = {'SVAL'}
 
+    autosave_defaults = make_autosave_defaults(
+        input_pass0=[
+            "ZSV",  # Zero Error Severity
+            "OSV",  # One Error Severity
+            "SIMS",  # Simulation Mode Severity
+        ],
+        output_pass0=[
+            "ZSV",  # Zero Error Severity
+            "OSV",  # One Error Severity
+            "COSV",  # Change of State Sevr
+            "SIMS",  # Simulation Mode Severity
+        ],
+    )
+
 
 class IntegerRecordPackage(TwincatTypeRecordPackage):
     """Create a set of records for an integer Twincat Variable"""
@@ -523,6 +649,38 @@ class IntegerRecordPackage(TwincatTypeRecordPackage):
     input_only_fields = {'AFTC', 'AFVL', 'SVAL'}
     dtyp = 'asynInt32'
 
+    autosave_defaults = make_autosave_defaults(
+        input_pass0=[
+            # Alarm severities
+            "HHSV",  # Hihi Severity
+            "HSV",  # High Severity
+            "LLSV",  # Lolo Severity
+            "LSV",  # Low Severity
+            "SIMS",  # Simulation Mode Severity
+            # Alarm limits
+            "HIHI",  # Hihi Alarm Limit
+            "LOLO",  # Lolo Alarm Limit
+            "HIGH",  # High Alarm Limit
+            "LOW",  # Low Alarm Limit
+        ],
+        output_pass0=[
+            # Alarm severities
+            "HHSV",  # Hihi Severity
+            "HSV",  # High Severity
+            "LLSV",  # Lolo Severity
+            "LSV",  # Low Severity
+            "SIMS",  # Simulation Mode Severity
+            # Control limits
+            "DRVH",  # Drive High Limit
+            "DRVL",  # Drive Low Limit
+            # Alarm limits
+            "HIHI",  # Hihi Alarm Limit
+            "LOLO",  # Lolo Alarm Limit
+            "HIGH",  # High Alarm Limit
+            "LOW",  # Low Alarm Limit
+        ],
+    )
+
 
 class FloatRecordPackage(TwincatTypeRecordPackage):
     """Create a set of records for a floating point Twincat Variable"""
@@ -535,12 +693,72 @@ class FloatRecordPackage(TwincatTypeRecordPackage):
                           'RBV'}
     input_only_fields = {'AFTC', 'AFVL', 'SMOO', 'SVAL'}
 
-    autosave_defaults = {
-        'input': dict(pass0={'PREC'},
-                      pass1={}),
-        'output': dict(pass0={'VAL', 'PREC'},
-                       pass1={}),
-    }
+    autosave_defaults = make_autosave_defaults(
+        input_pass0=[
+            # Basics
+            "PREC",  # precision
+            # Alarm severities
+            "HHSV",  # Hihi Severity
+            "HSV",  # High Severity
+            "LLSV",  # Lolo Severity
+            "LSV",  # Low Severity
+            "SIMS",  # Simulation Mode Severity
+            # Alarm limits
+            "HIHI",  # Hihi Alarm Limit
+            "LOLO",  # Lolo Alarm Limit
+            "HIGH",  # High Alarm Limit
+            "LOW",  # Low Alarm Limit
+        ],
+        output_pass0=[
+            # Basics
+            "PREC",  # precision
+            # Alarm severities
+            "HHSV",  # Hihi Severity
+            "HSV",  # High Severity
+            "LLSV",  # Lolo Severity
+            "LSV",  # Low Severity
+            "SIMS",  # Simulation Mode Severity
+            # Control limits
+            "DRVH",  # Drive High Limit
+            "DRVL",  # Drive Low Limit
+            # Alarm limits
+            "HIHI",  # Hihi Alarm Limit
+            "LOLO",  # Lolo Alarm Limit
+            "HIGH",  # High Alarm Limit
+            "LOW",  # Low Alarm Limit
+        ],
+    )
+
+    def get_scale_offset(self):
+        """Get the scale and offset for the analog record(s)."""
+        scale = self.config.get("scale", None)
+        offset = self.config.get("offset", None)
+        if scale is None and offset is None:
+            return None, None
+
+        scale = "1.0" if scale is None else scale
+        offset = offset or "0.0"
+        return scale, offset
+
+    def generate_input_record(self):
+        record = super().generate_input_record()
+        scale, offset = self.get_scale_offset()
+        if scale is not None:
+            # If LINR==SLOPE, VAL = RVAL * ESLO + EOFF
+            record.fields["LINR"] = "SLOPE"
+            record.fields["ESLO"] = scale
+            record.fields["EOFF"] = offset
+        return record
+
+    def generate_output_record(self):
+        record = super().generate_output_record()
+        scale, offset = self.get_scale_offset()
+        if scale is not None:
+            # If LINR==SLOPE, then RVAL = (VAL - EOFF) / ESLO
+            record.fields["LINR"] = "SLOPE"
+            record.fields["ESLO"] = scale
+            record.fields["EOFF"] = offset
+        return record
 
 
 class EnumRecordPackage(TwincatTypeRecordPackage):
@@ -550,6 +768,51 @@ class EnumRecordPackage(TwincatTypeRecordPackage):
     dtyp = 'asynInt32'
     output_only_fields = {'DOL', 'IVOA', 'IVOV', 'OMSL', 'ORBV', 'RBV'}
     input_only_fields = {'AFTC', 'AFVL', 'SVAL'}
+    autosave_defaults = make_autosave_defaults(
+        input_pass0=[
+            # Per-state severities
+            "ZRSV",  # State Zero Severity
+            "ONSV",  # State One Severity
+            "TWSV",  # State Two Severity
+            "THSV",  # State Three Severity
+            "FRSV",  # State Four Severity
+            "FVSV",  # State Five Severity
+            "SXSV",  # State Six Severity
+            "SVSV",  # State Seven Severity
+            "EISV",  # State Eight Severity
+            "NISV",  # State Nine Severity
+            "TESV",  # State Ten Severity
+            "ELSV",  # State Eleven Severity
+            "TVSV",  # State Twelve Severity
+            "TTSV",  # State Thirteen Sevr
+            "FTSV",  # State Fourteen Sevr
+            "FFSV",  # State Fifteen Severity
+            "UNSV",  # Unknown State Severity
+            "SIMS",  # Simulation Mode Severity
+        ],
+        output_pass0=[
+            # Per-state severities
+            "ZRSV",  # State Zero Severity
+            "ONSV",  # State One Severity
+            "TWSV",  # State Two Severity
+            "THSV",  # State Three Severity
+            "FRSV",  # State Four Severity
+            "FVSV",  # State Five Severity
+            "SXSV",  # State Six Severity
+            "SVSV",  # State Seven Severity
+            "EISV",  # State Eight Severity
+            "NISV",  # State Nine Severity
+            "TESV",  # State Ten Severity
+            "ELSV",  # State Eleven Severity
+            "TVSV",  # State Twelve Severity
+            "TTSV",  # State Thirteen Sevr
+            "FTSV",  # State Fourteen Sevr
+            "FFSV",  # State Fifteen Sevr
+            "UNSV",  # Unknown State Sevr
+            "COSV",  # Change of State Sevr
+            "SIMS",  # Simulation Mode Severity
+        ],
+    )
 
     mbb_fields = [
         ('ZRVL', 'ZRST'),
@@ -586,6 +849,18 @@ class WaveformRecordPackage(TwincatTypeRecordPackage):
     """Create a set of records for a Twincat Array"""
     input_rtyp = 'waveform'
     output_rtyp = 'waveform'
+    field_defaults = {
+        'APST': 'On Change',
+        'MPST': 'On Change',
+    }
+    autosave_defaults = make_autosave_defaults(
+        input_pass0=[
+            "SIMS",  # Simulation Mode Severity
+        ],
+        output_pass0=[
+            "SIMS",  # Simulation Mode Severity
+        ],
+    )
 
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
@@ -664,7 +939,33 @@ class StringRecordPackage(TwincatTypeRecordPackage):
     input_rtyp = 'waveform'
     output_rtyp = 'waveform'
     dtyp = 'asynInt8'
-    field_defaults = {'FTVL': 'CHAR'}
+    field_defaults = {
+        'FTVL': 'CHAR',
+        'APST': 'On Change',
+        'MPST': 'On Change',
+    }
+    autosave_defaults = make_autosave_defaults(
+        input_pass0=[
+            "SIMS",  # Simulation Mode Severity
+        ],
+        output_pass0=[
+            "SIMS",  # Simulation Mode Severity
+        ],
+    )
+    # The LSO record that gets generated when a "link" is used to write
+    # string data into the PLC from another EPICS PV:
+    autosave_defaults_for_link = make_autosave_defaults(
+        input_pass0=[
+            "SIMS",  # Simulation Mode Severity
+        ],
+        output_pass0=[
+            "SIMS",  # Simulation Mode Severity
+        ],
+    )
+
+    # Links to string PVs require auxiliary 'lso' record.
+    link_requires_record = True
+    link_suffix = "LSO"
 
     @property
     def nelm(self):
@@ -685,6 +986,39 @@ class StringRecordPackage(TwincatTypeRecordPackage):
         record.fields['NELM'] = self.nelm
         return record
 
+    def generate_link_record(self):
+        """An auxiliary 'lso' link record to pass string PVs to the PLC."""
+        record = EPICSRecord(
+            self.delimiter.join((self.pvname, self.link_suffix)),
+            record_type="lso",
+            direction="output",
+            package=self,
+            autosave=self.autosave_defaults_for_link["output"],
+        )
+        record.fields.pop('TSE', None)
+        record.fields.pop('PINI', None)
+        record.fields["SIZV"] = self.nelm
+
+        # Add our port
+        record.fields.update(
+            self._get_omsl_fields()
+        )
+        record.fields['OUT'] = f"{self.pvname} PP MS"
+        _update_description(record, f"Aux link record for {self.chain.tcname}")
+        return record
+
+    @property
+    def records(self):
+        """All records that will be created in the package"""
+        records = [self.generate_input_record()]
+        link_fields = self._get_omsl_fields()
+        if self.io_direction == 'output' or link_fields:
+            records.append(self.generate_output_record())
+            if link_fields and self.link_requires_record:
+                records.append(self.generate_link_record())
+
+        return records
+
 
 DATA_TYPES = {
     'BOOL': BinaryRecordPackage,
